import java.awt.*;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.DropTargetEvent;
import java.awt.dnd.DropTargetListener;
import java.awt.event.*;
import java.io.*;
import java.util.Arrays;
import ij.measure.Measurements;
import ij.measure.ResultsTable;
import ij.plugin.filter.Analyzer;
import ij.plugin.frame.*;
import ij.*;
import ij.process.*;
import ij.gui.*;
import ij.io.OpenDialog;
import ij.io.DirectoryChooser;
/**
 * Helper plug-in for Isotope Microscope.
 * @author Koutaro Abe ( kabe@cris.hokudai.ac.jp )
 * @version 10-Aug-2010
 */
@SuppressWarnings("serial")
public class ImsVIEW_ extends PlugInFrame
	implements ActionListener, ItemListener, 
	MouseListener, MouseWheelListener, KeyListener, DropTargetListener {

	static final String version="0810-2010";
//	private static int rows = 15;
//	private static int lastNonShiftClick = -1;
//	private static boolean allowMultipleSelections = true; 
//	private java.awt.List imglist;
	private Panel mainBtPanel;
	private PopupMenu substPopup;
	private Button substButton; 
	private static String subtButtonLabel = "Subtract "+'\u00bb';
	int previousID;
	static Frame instance;
	String[] mainButtons = {
			"Open as Stack", "Open separately",
			subtButtonLabel,
			"Line", "Area", "Depth", "Measure all",
			"ROI Manager", "Sync ROI",
			"Inspect" };
	DropTarget openBtn, openSliceBtn, openHDFBtn;

	public ImsVIEW_() {
		super("ImsVIEW");
		if ( IJ.versionLessThan( "1.39t" ) ){
			IJ.showMessage( "This plug-in requires ImageJ version 1.39t or later." );
			return;
		}
		if ( instance!=null ) {
//			IJ.showMessage( "Another ImsVIEW running." );
			instance.setVisible( true );
			instance.toFront();
			return;
		}
		instance = this;
//		imglist = new List( rows, allowMultipleSelections );
		showWindow();
	}
	
	void showWindow() {
		ImageJ ij = IJ.getInstance();
 		addKeyListener(ij);
 		addMouseListener(this);
		addMouseWheelListener(this);
		WindowManager.addWindow(this);

		//setLayout(new FlowLayout(FlowLayout.CENTER,5,5)); 
		setLayout(new BorderLayout());
		this.setTitle( "ImsVIEW ver."+version );
/*		imglist.add("012345678901234");
		imglist.addItemListener(this);
		imglist.addKeyListener(ij);
		imglist.addMouseListener(this);
		imglist.addMouseWheelListener(this); 
		if (IJ.isLinux()) imglist.setBackground(Color.white);
		add( "Center", imglist );*/
		
		mainBtPanel = new Panel();
		addButtons( mainBtPanel, mainButtons );
		add( mainBtPanel );
		addPopupMenu();

		pack();
		GUI.center(this);
		Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
		instance.setLocation( screen.width - instance.getWidth() - 16 , 16 );
		setVisible(true);
	}
	
	void addButtons( Panel panel, String[] buttons ) {
		panel.setLayout( new GridLayout( buttons.length, 1, 4, 4 ) );
		for( int i=0; i<buttons.length; i++ )
			addButton( panel, buttons[i] );
	}
	void addButton( Panel panel, String label ) {
		Button b = new Button( label );
//		b.setPreferredSize(null);
		b.addActionListener( this );
		b.addKeyListener( IJ.getInstance() );
 		if (label.equals(subtButtonLabel)) substButton = b;
 		panel.add( b );
 		if( label.equals( mainButtons[0] ) )
 			openBtn = new DropTarget( b, this );
 		if( label.equals( mainButtons[1] ) )
 			openSliceBtn = new DropTarget( b, this );
 //		if( label.equals( mainButtons[2] ) )
 //			openHDFBtn = new DropTarget( b, this );
	}
	void addPopupMenu() {
		substPopup = new PopupMenu();
		//addPopupItem( pm, "Select All" );
		addPopupItem( substPopup, "Total image" );
		addPopupItem( substPopup, "Total image /slice" );
		addPopupItem( substPopup, "Integral stack" );
		addPopupItem( substPopup, "Differential stack" );
//		addPopupItem( substPopup, "BATCH Total image /slice" );
		add( substPopup );
	}
	void addPopupItem( PopupMenu pm, String s ) {
		MenuItem mi=new MenuItem( s );
		mi.addActionListener( this );
		pm.add( mi );
	}

	public void actionPerformed( ActionEvent e ) {
//		int modifiers = e.getModifiers();
//		boolean altKeyDown = (modifiers&ActionEvent.ALT_MASK)!=0 || IJ.altKeyDown();
//		boolean shiftKeyDown = (modifiers&ActionEvent.SHIFT_MASK)!=0 || IJ.shiftKeyDown();
		IJ.setKeyUp(KeyEvent.VK_ALT);
		IJ.setKeyUp(KeyEvent.VK_SHIFT);
		String label = e.getActionCommand();
		if( label==null )
			return;
		else if (label.equals(subtButtonLabel) ) {
			Point ploc = mainBtPanel.getLocation();
			Point bloc = substButton.getLocation();
			substPopup.show( this, ploc.x, bloc.y );
		} 
		ImagePlus imp = WindowManager.getCurrentImage();
		if( imp!=null ) {
			ImageProcessor ip = imp.getProcessor();
			int id = imp.getID();
			if( id!=previousID )
				ip.snapshot();
			previousID = id;
			new ImsVIEWRunner( label, imp, ip );
		} else {
			IJ.showStatus( "No image" );
			previousID = 0;
			new ImsVIEWRunner( label );
		}
	}

	@Override
	public void processWindowEvent(WindowEvent e) {
		super.processWindowEvent(e);
		if( e.getID()==WindowEvent.WINDOW_CLOSING )
			instance = null;	
	}
	public void itemStateChanged(ItemEvent arg0) { }
	public void mouseClicked(MouseEvent arg0) { }
	public void mouseEntered(MouseEvent arg0) { }
	public void mouseExited(MouseEvent arg0) { }
	public void mousePressed(MouseEvent arg0) { }
	public void mouseReleased(MouseEvent arg0) { }
	public void mouseWheelMoved(MouseWheelEvent arg0) { }
	public void keyPressed(KeyEvent arg0) { }
	public void keyReleased(KeyEvent arg0) { }
	public void keyTyped(KeyEvent arg0) { }

	public void dragEnter(DropTargetDragEvent arg0) { }
	public void dragExit(DropTargetEvent arg0) { }
	public void dragOver(DropTargetDragEvent arg0) { }
	@SuppressWarnings("unchecked")
	public void drop( DropTargetDropEvent dtde ) {
		// based on DropTest2.java 
		try {
			// Ok, get the dropped object and try to figure out what it is
			Transferable tr = dtde.getTransferable();
			DataFlavor[] flavors = tr.getTransferDataFlavors();
			for (int i = 0; i < flavors.length; i++) {
//				IJ.log( "Possible flavor: "	+ flavors[i].getMimeType());
				// Check for file lists specifically
				if (flavors[i].isFlavorJavaFileListType()) {
					// Great! Accept copy drops...
					dtde.acceptDrop(DnDConstants.ACTION_COPY_OR_MOVE);
					// IJ.log( "Successful file list drop.\n\n" );

					// And add the list of file names to our text area
					java.util.List<File> list = (java.util.List) tr.getTransferData(flavors[i]);
//					for (int j = 0; j < list.size(); j++)
//						IJ.log( list.get(j) +":" + list.get(j).getClass()+"\n" );
					if( dtde.getSource().equals( openBtn ) ) { new ImsVIEWRunner( "Drop:Open as Stack", list ); }
					if( dtde.getSource().equals( openSliceBtn ) ) { new ImsVIEWRunner( "Drop:Open separately", list ); }
//					if( dtde.getSource().equals( openHDFBtn ) ) { new ImsVIEWRunner( "Drop:Open HDF v3.2", list ); }

					// If we made it this far, everything worked.
					dtde.dropComplete(true);
					return;
				}
			}
			// Hmm, the user must not have dropped a file list
			IJ.log( "Drop failed: " + dtde );
			dtde.rejectDrop();
		} catch (Exception e) {
			e.printStackTrace();
			dtde.rejectDrop();
		}
	}
	public void dropActionChanged(DropTargetDragEvent arg0) { }
}

class ImsVIEWRunner extends Thread {

	static final String DIR_OPENSTACK = "ims.openAsStack.defaultdir";
	static final String DIR_OPENSLICE = "ims.openAsSlice.defaultdir";

	private String command;
	private ImagePlus imp;
	private ImageProcessor ip;
	private java.util.List<File> target;
	RoiManager roiManager;
	ResultsTable resultTable;

	ImsVIEWRunner( String command, ImagePlus imp, ImageProcessor ip, java.util.List<File> target ) {
		super(command);
		this.command = command;
		this.imp = imp;
		this.ip = ip;
		this.target = target;
		setPriority(Math.max(getPriority()-2, MIN_PRIORITY));
		start();
	}

	ImsVIEWRunner( String command, ImagePlus imp, ImageProcessor ip ) { this( command, imp, ip, null ); }
	ImsVIEWRunner( String command, java.util.List<File> target ) { this( command, null, null, target ); }
	ImsVIEWRunner( String command ) { this( command, null, null, null ); }

	@Override
	public void run() {
		try {
			runCommand( command, imp, ip, target );
		} catch ( OutOfMemoryError e ) {
			IJ.outOfMemory( command );
		} catch( Exception e ) {
			CharArrayWriter caw = new CharArrayWriter( );
			PrintWriter pw = new PrintWriter( caw );
			e.printStackTrace( pw );
			IJ.log( caw.toString( ) );
			IJ.showStatus("");
		}
		if ( imp!=null ) imp.unlock();
	}

	void runCommand(String command, ImagePlus imp, ImageProcessor ip, java.util.List<File> target ) {
		IJ.showStatus(command + "...");
		if( imp!=null && ip!=null ) { // requires current image
			if( command.equals( "Delete Current Slice" ) ) { 
				ImageStack s = checkStack(imp);
				if ( s==null )
					return;
				int n = imp.getCurrentSlice();
				s.deleteSlice( n );
				if (s.getSize()==1) {
					imp.setProcessor( null, s.getProcessor(1) );
		 			new ImageWindow( imp );
				}
				imp.setStack( s );
		        imp.unlock();				
			}
			else if ( command.equals("Integral stack") ) {
				if ( !checkImp(imp) )
					return;
				Delta delta = new Delta( imp, "integral" );
		        imp.unlock();	
				delta.show();
			}
			else if ( command.equals("Differential stack") ) {
				if ( !checkImp(imp) )
					return;
				Delta delta = new Delta( imp, "each" );
		        imp.unlock();	
				delta.show();
			}
			else if ( command.equals("Total image" ) ) {
				if ( !checkImp(imp) )
					return;
				Delta delta = new Delta( imp, "total" );
		        imp.unlock();	
				delta.show();
			}
			else if ( command.equals("Total image /slice" ) ) {
				if ( !checkImp(imp) )
					return;
				Delta delta = new Delta( imp, "total+div" );
		        imp.unlock();	
				delta.show();
			}
			else if( command.equals("Depth") ) { ijRun("Plot Z-axis Profile"); }
			else if( command.equals("Area") ) { ijRun("Measure"); }
			else if( command.equals("Line") ) { ijRun("Dynamic Profiler"); }
			else if( command.equals("Sync ROI") ) { displayRois(); }
			else if( command.equals("Inspect") ) { new PixelInspector_(); }
			else if( command.equals("Measure all") ) {
				int[] image = WindowManager.getIDList();
//				IJ.log( "imageId count:"+imageId.length + ", getImageCount" + WindowManager.getImageCount() );
				if( image.length<1 ) // no image windows opened
					return;
				if( resultTable==null )
					resultTable = new ResultsTable();
				Roi[] rois = { null };
				roiManager = RoiManager.getInstance();
				if( roiManager!=null )
					rois = roiManager.getSelectedRoisAsArray();
				for( int j=0; j<rois.length; j++ )
					for( int i=0; i<image.length; i++ ){
//						IJ.log( "image: "+image[i]+", roi:"+rois[j] );
						ImagePlus currentImp = WindowManager.getImage( image[i] );
						if( currentImp==null ) continue;
						measureImp( currentImp, rois[j], resultTable  );
					}
				resultTable.show("Results all images");
			}
		}
		// current image not required 
		if ( command.equals("Open as Stack") ){
			DirectoryChooser.setDefaultDirectory( Prefs.get( DIR_OPENSTACK, Prefs.getDefaultDirectory() ) );
			DirectoryChooser od = new DirectoryChooser( "Open directory for '32bit-Real Raw' or 'HDF v3.2' ..." );
			if( od!=null && od.getDirectory()!=null && od.getDirectory()!="" ) {
				ImsOpener d = new ImsOpener( od.getDirectory(), false, true );
				d.open();
				Prefs.set( DIR_OPENSTACK, od.getDirectory() );
				setBCWindow();
			}

		} else if( command.equals("Drop:Open as Stack") ){
			ImsOpener d = new ImsOpener( target, false, true );
			d.open();
			setBCWindow();

		} else if( command.equals("Open separately") ){
			OpenDialog.setDefaultDirectory( Prefs.get( DIR_OPENSLICE, Prefs.getDefaultDirectory() ) );
			OpenDialog od = new OpenDialog( "Open file '32bit-Real Raw' or 'HDF v3.2' ...", null );			
			if( od!=null && od.getDirectory()!=null && od.getDirectory()!="" ) {
				ImsOpener d = new ImsOpener( od.getDirectory()+od.getFileName(), false, false );
				d.open();
				setBCWindow();
				Prefs.set( DIR_OPENSLICE, od.getDirectory() );
			}

		} else if( command.equals("Drop:Open separately") ){
			ImsOpener d = new ImsOpener( target, false, false );
			d.open();
			setBCWindow();

		} else if( command.equals("Open HDF v3.2") ) {
			OpenDialog od = new OpenDialog( "Open HDF v3.2 file...", null );
			new HDF32Opener( od.getDirectory()+od.getFileName() ).open(true);
			setBCWindow();

		} else if( command.equals("Drop:Open HDF v3.2") ) {
			for( int i=0; i<target.size(); i++ )
				if( target.get(i).isFile() )
					new HDF32Opener( target.get(i).toString() ).open(true);
			setBCWindow();

		} else if( command.equals("BATCH Total image /slice") ) {
			try {
				String topDir;
				ImagePlus sourceImp, destImp;
				topDir = new DirectoryChooser( "Select top folder" ).getDirectory();
				if( topDir==null ) return;
				java.util.List<File> list = Arrays.asList( new File(topDir).listFiles() );	// fBNg̃t@CXg
				IJ.log( "BATCH - Top directory:"+topDir );
				for( int i=0; i<list.size(); i++ ) {
					// fBNg̃t@C(source)Ƃloop
					if( !list.get(i).isDirectory() ) continue;
					// fBNg̃fBNgAX^bN摜ƂĊJ
					ImsOpener d = new ImsOpener( list.subList(i,i), false, true );
					d.open();
					sourceImp = WindowManager.getCurrentImage();
					if ( sourceImp==null || sourceImp.getStack().getSize()<2 ) continue;
					// V摜iDeltaj
					Delta delta = new Delta( sourceImp, "total+div" );
					delta.show();
					destImp = WindowManager.getCurrentImage();
					String saveFilename = delta.getTitle()+".tif";
					// V摜iDeltajۑ
					Macro.saveAs( topDir+"/"+saveFilename );
					IJ.log( "Saved  '"+saveFilename+"'." );
					destImp.close();
					//@
					if( sourceImp!=null ) sourceImp.close();
				}
				IJ.log( "BATCH - complete." );
			} catch( Exception e ) {
				if (!Macro.MACRO_CANCELED.equals(e.getMessage()))
					IJ.showMessage( ""+e );
				return;
			}
		} else if ( command.equals("ROI Manager") ) {
			roiManager = RoiManager.getInstance();
			if( roiManager==null )
				roiManager = new RoiManager();
			roiManager.toFront();
		}

		if( imp!=null ) {
			imp.updateAndDraw();
			imp.unlock();
		}
		IJ.showStatus("");
	}

	void setBCWindow() { setBCWindow(0.35); } 
	void setBCWindow( double saturated ) {
		try {
			IJ.run("Brightness/Contrast...");
			IJ.run("Enhance Contrast", "saturated="+Double.toString( saturated ) );
		} catch( Exception e ) {
			if (!Macro.MACRO_CANCELED.equals(e.getMessage()))
				IJ.handleException(e);			
		}
	}
	void ijRun( String Command ) {
		try {
			IJ.run( Command );				
		} catch( Exception e ) {
			if (!Macro.MACRO_CANCELED.equals(e.getMessage()))
				IJ.handleException(e);
		}		
	}
	boolean checkImp( ImagePlus imp ) {
		if ( imp==null ) {
			IJ.showMessage( "Please select a image window.");
			return false;
		}
		if ( !imp.lock() ) {
			IJ.showMessage( "Could not lock image.");
			return false;
		}
		return true;
	}

	ImageStack checkStack( ImagePlus imp ){
		if( checkImp( imp )==false )
			return null;
		ImageStack s = imp.getStack();
		if ( s.getSize()<2 ) {
			IJ.showMessage( "Selected image is not a Stack.");
			return null;
		}
		return s;
	}

	boolean openHDF32( String filepath ) {
		try {
			if ( filepath==null )
				return false;
			new HDF32Opener( filepath );
			return true;
		} catch( Exception e ) {
			if (!Macro.MACRO_CANCELED.equals(e.getMessage()))
				IJ.showMessage( ""+e );
			return false;
		}
	}

	void displayRois() {
		Roi roi = imp.getRoi();
		if( roi==null )
			return;
		int[] imageId = WindowManager.getIDList();
		for( int j=0; j<imageId.length; j++ ){
			ImagePlus imp = WindowManager.getImage( imageId[j] );
			if( imp==null )
				continue;
			imp.setRoi( roi, true );
			ImageCanvas ic = imp.getCanvas();
			if( ic==null )
				continue;
			ic.setShowAllROIs(true);
			imp.draw();
		}
	}

	void displayValue() {
		ResultsTable rt = new ResultsTable();
		ImageProcessor ip = imp.getProcessor();
		float[][] f = ip.getFloatArray();
		Roi roi = imp.getRoi();
		Rectangle r = roi.getBounds();
//		for( int x=(int) r.getMinX(); x<r.getMaxX(); x++ ) {		rt.addColumns( );
//		}
		for( int y=(int) r.getMinY(); y<r.getMaxY(); y++ ) {
			rt.incrementCounter();
			for( int x=(int) r.getMinX(); x<r.getMaxX(); x++ ) {
				rt.addValue( x-(int) r.getMinX(), f[x][y] );
			}
		}
	}
	boolean measureAllImages( ) {
		ResultsTable rt = new ResultsTable();
		rt.show("Results all images");
		int[] imageId = WindowManager.getIDList();
		IJ.log( "imageId count:"+imageId.length + ", getImageCount" + WindowManager.getImageCount() );
		for( int i=0; i<imageId.length; i++ ){
//			IJ.log( "imageId:"+i+", "+imageId[i] );
			ImagePlus imp = WindowManager.getImage( imageId[i] );
			if( imp==null )	continue;
			IJ.log( "  run measure" );
//			measureImp( imp, rt );
		}
		rt.show("Results all images");
		return true;
	}
	boolean measureImp(ImagePlus imp, Roi roi, ResultsTable rt ) {
		if (imp==null)
			return false;

		int measurements = Analyzer.getMeasurements();
		if( imp.getStackSize()>1 )
			Analyzer.setMeasurements(measurements|Measurements.SLICE);
		
		if( roi!=null )
			imp.setRoi( roi );
		Analyzer analyzer = new Analyzer(imp, measurements, rt);
		analyzer.measure();
//		IJ.run( "Measure" );
		return true;
	}	

	ImagePlus getImage() {
		ImagePlus imp = WindowManager.getCurrentImage();
		if (imp==null) {
			error("There are no images open.");
			return null;
		} else
			return imp;
	}
	ImagePlus getImage( int id ) {
		ImagePlus imp = WindowManager.getImage( id );
		if (imp==null) {
			error("There are no images open.");
			return null;
		} else
			return imp;
	}
	void continuosConv() { }
	boolean error(String msg) {
		IJ.showMessage(msg);
		Macro.abort();
		return false;
	}
}
